#include <Autolock.h>
#include <Dragger.h>
#include <PopUpMenu.h>
#include <Roster.h>
#include <MenuItem.h>

#include "ResourceUtils.h"

#include "ConfigView.h"
#include "MountainRenderer.h"
#include "MountainView.h"

#define CONFIGURE_MSG 'Cnfg'

extern char* app_signature;

/*----------------------------------------------------------------------------------------
	° MountainView
------------------------------------------------------------------------------------------*/

BArchivable* MountainView::Instantiate(BMessage* archive)
{	
	if(validate_instantiation(archive,"MountainView"))
		return new MountainView(archive);
	return NULL;

}

status_t MountainView::Archive(BMessage* archive,bool deep) const
{	
	BitmapView::Archive(archive,deep);
	
	archive->AddString("class","MountainView");
	archive->AddString("add_on", app_signature);
	archive->AddData("mRenderer",B_RAW_TYPE,mRenderer,sizeof(MountainRenderer));
	
	return B_OK;
}

MountainView::MountainView(BMessage* archive) : BitmapView(archive)
{
	mAllDone = true;
	mThread = -1;
	mRenderer = new MountainRenderer;
	mViewFilter = NULL;
	int32 size;
	const void * s = NULL;
	archive->FindData("mRenderer",B_RAW_TYPE,&s,&size);
	if(s)
		memcpy(mRenderer,s,sizeof(MountainRenderer));
}

MountainView::MountainView(BRect frame, const char* name,
							uint32 mode,uint32 flags)
			: BitmapView(frame,name,mode,flags)
{
	mAllDone = true;
	mThread = -1;
	mRenderer = new MountainRenderer;
	mViewFilter = NULL;
}

MountainView::~MountainView()
{	
	thread_info info;
	get_thread_info(mThread,&info);
	
	mAllDone = true;
	int32 val;

	if(info.state == B_THREAD_ASLEEP)
		kill_thread(mThread);
	else
		wait_for_thread(mThread,&val);

	mThread = -1;
	
	delete mRenderer;
	
	if(mViewFilter)
		mViewFilter->mMountainView = NULL;

}

/*----------------------------------------------------------------------------------------
	° AttachedToWindow
------------------------------------------------------------------------------------------*/

void MountainView::AttachedToWindow()
{	
	BDragger* dragger = cast_as(ChildAt(0),BDragger);
	if(dragger != NULL)
	{
		dragger->PopUp()->ItemAt(3)->SetTarget(this);
	}
	if(mAllDone)
		Start();
}

/*----------------------------------------------------------------------------------------
	° MessageReceived
------------------------------------------------------------------------------------------*/

void MountainView::MessageReceived(BMessage* msg)
{
	switch(msg->what)
	{
		case CONFIGURE_MSG:
		{
			if(mViewFilter == NULL)
			{
				entry_ref ref;
				be_roster->FindApp(app_signature,&ref);
				BFile file(&ref, B_READ_ONLY);
				BResources res(&file);
						
				BWindow* win = cast_as(ResourceUtils::RestoreFromResource(&res, 'ARCV', 1), BWindow);
				if(win != NULL)
				{
					MountainConfigView* view = cast_as(win->FindView("ConfigView"),MountainConfigView);
					view->mQuitApp = false;
					view->SyncControls(mRenderer);
					
					mViewFilter = new ConfigViewFilter();
					mViewFilter->mMountainView = this;
					win->AddCommonFilter(mViewFilter);
					win->Show();
				}
			}
		}		
		break;
		
		default:
			BitmapView::MessageReceived(msg);
		break;
	}
}

/*----------------------------------------------------------------------------------------
	° ChangeSettings
------------------------------------------------------------------------------------------*/

void MountainView::ChangeSettings(MountainRenderer* r)
{	
	if(mThread > 0)
	{
		thread_info info;
		get_thread_info(mThread,&info);
		
		mAllDone = true;
		if(info.state == B_THREAD_ASLEEP)
			kill_thread(mThread);
	
		int32 val;
		wait_for_thread(mThread,&val);
		mThread = -1;
	}
	
	*mRenderer = *r;
	{
		BWindow* win = Window();
		BAutolock lock(win);
		
		if(win->Bounds() == Bounds())
			win->ResizeTo(r->s_width, r->s_height);
	
		ResizeTo(r->s_width, r->s_height);	
	}
	BRect rect(0,0,r->s_width - 1, r->s_height - 1);
	BBitmap* image = new BBitmap(rect,B_COLOR_8_BIT,true);
	image->AddChild(new BView(rect,"",0,0));
	SetImage(image);
	
	Start();
}

/*----------------------------------------------------------------------------------------
	° Start
------------------------------------------------------------------------------------------*/

void MountainView::Start()
{	
	if(mThread > 0) return;
	
	mAllDone = false;

	mThread = spawn_thread(thread_func, "Mountains", B_LOW_PRIORITY, this);
	if(mThread > 0)
		resume_thread(mThread);
}

/*----------------------------------------------------------------------------------------
	° thread_func
------------------------------------------------------------------------------------------*/

int32 MountainView::thread_func(void * data)
{
	MountainView* view = (MountainView*)data;
	
	BBitmap* image = view->mImage;
	if(image == NULL) return -1;
	
	MountainRenderer* r = view->mRenderer;
	r->init_graphics();

	for(;;)
	{		
		{
			BAutolock lock(view->mBitmapLock);
			r->PlotNextColumn(image);
		}		
		{
			BAutolock lock(view->Window());		
			if(r->mScrolling)
			{	
				if(!r->repeat) break;
				view->Invalidate();
			}else
			{		
				BRect rect;
				rect.top = 0; rect.bottom = r->s_height;
				// rather than sync drawing, we'll just invalidate a bigger rect.
				rect.left = r->mCurCol - 5;
				rect.right =  r->mCurCol;
				view->Invalidate(rect);
			}
		}
		if(view->mAllDone) break;
		if(r->mScrolling)
			snooze(r->snooze_time * 1000);
		if(view->mAllDone) break;
	}
	
	view->mThread = -1;
	
	return 0;
}

#pragma mark -
#pragma mark === ConfigViewFilter ===
/*----------------------------------------------------------------------------------------
	° ConfigViewFilter
------------------------------------------------------------------------------------------*/

ConfigViewFilter::ConfigViewFilter() : BMessageFilter('Go  ')
{
	mMountainView = NULL;
}

ConfigViewFilter::~ConfigViewFilter()
{
	if(mMountainView)
		mMountainView->mViewFilter = NULL;
}

/*----------------------------------------------------------------------------------------
	° Filter
------------------------------------------------------------------------------------------*/

filter_result ConfigViewFilter::Filter(BMessage *msg, BHandler **target)
{
	MountainConfigView* view = cast_as(*target,MountainConfigView);
	if(view)
	{
		MountainRenderer* r = new MountainRenderer;
		view->Settings(r);
		
		if(mMountainView == NULL)
		{
			BRect frame(0, 0, r->s_width - 1, r->s_height - 1);
			
			BWindow* win = new BWindow(frame, "BeMountains", B_DOCUMENT_WINDOW, B_NOT_RESIZABLE);	
			mMountainView = new MountainView(frame,"MountainView",B_FOLLOW_NONE, B_WILL_DRAW);
			mMountainView->mViewFilter = this;
			
			// dragger
			frame.right = frame.left  + 7;
			frame.bottom= frame.top + 7;
			BDragger* dragger = new BDragger(frame, mMountainView);
			
			dragger->PopUp()->AddItem(new BMenuItem("Configure",new BMessage(CONFIGURE_MSG)));
			dragger->SetResizingMode(B_FOLLOW_NONE);
			
			// set up window
			mMountainView->AddChild(dragger);
			
			win->AddChild(mMountainView);
			win->MoveTo(100,100);			
			win->Show();
		}
		mMountainView->ChangeSettings(r);
		delete r;
	}
	return B_DISPATCH_MESSAGE;
}
